"""
FixedIncome
===========
Provides tools for Fixed Income analytics

This package was developed for 15.438 Fixed Income Securities and Derivatives at the MIT Sloan School of Management. Its recommended usage is as
    >>> import fixedincome as fi
Documentation is available through the built-in ``help`` function.

Available subpackages
---------------------
    utils
        Basic functions useful for fixed income applications.
    bonds
        Functions useful for analyzing fixed rate bonds.
"""

__all__ = ['utils', 'bonds']

from fixedincome import *



"""
Fixed income analytics (general)

Implements various generic functions useful for fixed income applications.
"""


import datetime

import numpy as np
import scipy.optimize


MONTHS_IN_YEAR = 12


def npv(rate_ : float, values : np.ndarray) -> float:
    """
    Calculates the Net Present Value (NPV) of a series of cash flows.

    Parameters
    ----------
    rate_ : float
        The (simple) discount rate over the length of one period.
    values : np.ndarray
        Correspond to payments and income per period, in order.

    Returns
    -------
    float
        The NPV of the given series of cash flows, discounted at the given rate.
    """

    weights = np.fromfunction(function=lambda i: 1/(1 + rate_)**(i + 1), shape=(values.size,), dtype=int)
    return np.dot(values, weights)


def irr(values: np.ndarray, guess : float = 0.1) -> float:
    """
    Calculates the Internal Rate of Return (IRR) for a series of cash flows.

    Parameters
    ----------
    values : np.ndarray
        A series of cash flows, in order.
    guess : float [optional]
        An initial guess for the IRR.

    Returns
    -------
    float
        The IRR of the given series of cash flows.
    """

    return scipy.optimize.newton(func=npv, x0=guess, args=(values,), tol=0.0000001, maxiter=20)


def rate(nper : int, pmt : float, pv_ : float, fv_ : float = 0, guess : float = 0.1) -> float:
    """
    Calculates the interest rate per period of an annuity.

    Parameters
    ----------
    nper : int
        The total number of payment periods in an annuity.
    pmt : float
        The constant payment amount made each period.
    pv_ : float
        The present value (i.e., total amount that a series of payments is worth now) of the annuity.
    fv_ : float [optional]
        The future value (i.e., cash balance to be attained after the last payment is made) of the annuity.
    guess : float [optional]
        An initial guess for the interest rate.

    Returns
    -------
    float
        The interest rate per period of the annuity.
    """

    cash_flows = np.array([pv_] + [pmt] * nper, dtype=float)
    cash_flows[-1] += fv_

    return irr(values=cash_flows, guess=guess)


def pv(rate_ : float, nper : int, pmt : float, fv_ : float = 0, type_ : int = 0) -> float:
    """
    Calculates the present value of a loan based on a constant interest rate.

    Parameters
    ----------
    rate_ : float
        The interest rate per period.
    nper : int
        The total number of payment periods in an annuity.
    pmt : float
        The constant payment amount made each period.
    fv_ : float [optional]
        The future value (i.e., cash balance to be attained after the last payment is made) of the annuity.
    type_ : int [optional]
        Indicates when payments are due.
            - 0 [default] : at the end of the period.
            - 1 : at the beginning of the period

    Returns
    -------
    float
        The present value of the loan based on a constant interest rate.
    """
    
    if rate != 0:
        _pv = -(pmt*(1 + rate_*type_)*(((1 + rate_)**nper - 1)/rate_) + fv_)/((1 + rate_)**nper)
    else:
        _pv = -(pmt*nper + fv_)

    return _pv


def fv(rate_ : float, nper : int, pmt : float, pv_ : float = 0, type_ : int = 0) -> float:
    """
    Calculates the future value of a loan based on a constant interest rate.

    Parameters
    ----------
    rate_ : float
        The interest rate per period.
    nper : int
        The total number of payment periods in an annuity.
    pmt : float
        The constant payment amount made each period.
    pv_ : float [optional]
        The present value of the annuity.
    type_ : int [optional]
        Indicates when payments are due.
            - 0 [default] : at the end of the period.
            - 1 : at the beginning of the period

    Returns
    -------
    float
        The future value of the loan based on a constant interest rate.
    """
    
    if rate_ != 0:
        _fv = -(pv_*(1 + rate_)**nper + pmt*(1 + rate_*type_)*(((1 + rate_)**nper - 1)/rate_))
    else:
        _fv = -(pmt*nper + pv_)

    return _fv


def _thirty_threesixty_day_count_factor(start : datetime.date, end : datetime.date) -> float:
    """Calculate day count factor on a 30/360 basis (after date adjustments)."""
    return (360 * (end.year - start.year) + 30 * (end.month - start.month) + (end.day - start.day)) / 360


def day_count_factor(start : datetime.date, end : datetime.date, basis : int = 0, next_ : datetime.date = None, freq : int = None) -> float:
    """
    Calculates the day count factor between dates for measuring interest accrual.

    Parameters
    ----------
    start : datetime.date
        The starting date of the interest period.
    end : datetime.date
        The ending date of the accrual period.
    basis : int [optional]
        The type of day count basis to use.
            - 0 [default] : US (NASD) 30/360
            - 1 : Actual/Actual
            - 2 : Actual/360
            - 3 : Actual/365
            - 4 : European 30/360
    next_ : datetime.date [optional, required when basis == 1]
        The ending date of the interest period, or starting date of the next interest period. This argument is required for the Actual/Actual basis.
    freq : int [optional, required when basis == 1]
        The number of interest payment periods in a full year.

    Returns
    -------
    float
        The day count factor, or percentage of interest that has linearly accrued, for the interest payment period.
    """

    # Switch calculation methodology by specified basis
    if basis == 0: # US (NASD) 30/360
        # Date adjustments
        if end.day == 31 and (start.day == 30 or start.day == 31):
            end = end.replace(day=30)
        if start.day == 31:
            start = start.replace(day=30)
        return _thirty_threesixty_day_count_factor(start=start, end=end)
    elif basis == 1: # Actual/Actual
        assert next_, "Valid date of next coupon payment required for Actual/Actual basis"
        assert freq and (freq > 0), "Valid number of annual coupon payments required for Actual/Actual basis"
        return (end - start).days / (freq * (next_ - start).days)
    elif basis == 2: # Actual/360
        return (end - start).days / 360.0
    elif basis == 3: # Actual/365
        return (end - start).days / 365.0
    elif basis == 4: # European 30/360
        # Date adjustments
        if start.day == 31:
            start = start.replace(day=30)
        if end.day == 31:
            end = end.replace(day=30)
        return _thirty_threesixty_day_count_factor(start=start, end=end)
    else: # Basis misspecified
        raise ValueError("Invalid basis specified.")



"""
Fixed income analytics for bonds

Implements various functions useful for bond analytics.
"""


import datetime
from typing import List

import dateutil.relativedelta
import numpy as np
import scipy.optimize

from .utils import MONTHS_IN_YEAR
from .utils import day_count_factor


def accrint(issue : datetime.date, first_interest : datetime.date, settlement : datetime.date, rate : float, par : float, frequency : int, basis : int = 0) -> float:
    """
    Returns the accrued interest for a bond.

    Parameters
    ----------
    issue : datetime.date
        The bond's issue date.
    first_interest : datetime.date
        The bonds's first coupon date.
    settlement : datetime.date
        The bond's settlement date.
    rate : float
        The bond's annual coupon rate.
    par : float
        The bond's par value.
    frequency : int
        The number of coupon payments per year.
    basis : int [optional]
        The type of day count basis to use.
            - 0 [default] : US (NASD) 30/360
            - 1 : Actual/Actual
            - 2 : Actual/360
            - 3 : Actual/365
            - 4 : European 30/360

    Returns
    -------
    float
        The accrued interest of the bond.
    """

    _day_count_factor = day_count_factor(start=issue, end=settlement, basis=basis, next_=first_interest, freq=frequency)

    # Use interest formula
    return par * rate * _day_count_factor


def coupon_dates(settlement : datetime.date, maturity : datetime.date, frequency : int) -> List[datetime.date]:
    """
    Returns the coupon dates for a bond from settlement to maturity.

    Parameters
    ----------
    settlement : datetime.date
        The bond's settlement date.
    maturity : datetime.date
        The bonds's maturity date (when it expires).
    frequency : int
        The number of coupon payments per year.

    Returns
    -------
    list[datetime.date]
        Coupon dates of the bond.
    """

    _coupon_dates = [maturity]

    # Calculate length of coupon period in months
    coupon_period = dateutil.relativedelta.relativedelta(months=-MONTHS_IN_YEAR//frequency)

    # Calculate coupon dates backwards from maturity to issuance
    while _coupon_dates[-1] > settlement:
        _coupon_dates.append(_coupon_dates[-1] + coupon_period)
    
    # Return list in chronological order
    _coupon_dates = _coupon_dates[::-1]
    return _coupon_dates


def price(settlement : datetime.date, maturity : datetime.date, rate : float, yld : float, redemption : float, frequency : int, basis : int = 0) -> float:
    """
    Returns the price per $100 face value of a bond.

    Parameters
    ----------
    settlement : datetime.date
        The bond's settlement date.
    maturity : datetime.date
        The bonds's maturity date (when it expires).
    rate : float
        The bond's annual coupon rate.
    yld : float
        The bond's annual yield.
    redemption : float
        The security's redemption value per $100 face value.
    frequency : int
        The number of coupon payments per year.
    basis : int [optional]
        The type of day count basis to use.
            - 0 [default] : US (NASD) 30/360
            - 1 : Actual/Actual
            - 2 : Actual/360
            - 3 : Actual/365
            - 4 : European 30/360

    Returns
    -------
    float
        Price per $100 face value of the bond.
    """

    _PAR = 100

    # Calculate coupon dates
    _coupon_dates = coupon_dates(settlement=settlement, maturity=maturity, frequency=frequency)
    num_periods = len(_coupon_dates) - 1 # First coupon date is before settlement

    # Calculate cash flows
    cash_flows = np.array([_PAR * rate / frequency] * num_periods) # Coupon payments
    cash_flows[-1] += redemption # Principal repayment

    # Calculate discount rates
    discount_rates = np.array([1 + yld / frequency] * num_periods)
    time_to_next = 1 - frequency * day_count_factor(start=_coupon_dates[0], end=settlement, basis=basis, next_=_coupon_dates[1], freq=frequency)
    discount_rate_powers = -1 * np.array([i + time_to_next for i in range(num_periods)])
    discount_factors = np.power(discount_rates, discount_rate_powers)

    # Calculate dirty price
    transaction_price = np.dot(cash_flows, discount_factors)

    # Calculate clean price
    _accrint = accrint(issue=_coupon_dates[0], first_interest=_coupon_dates[1], settlement=settlement, rate=rate, par=_PAR, frequency=frequency, basis=basis)
    clean_price = transaction_price - _accrint
    
    return clean_price


def yield_(settlement : datetime.date, maturity : datetime.date, rate : float, price_ : float, redemption : float, frequency : int, basis : int = 0):
    """
    Returns the yield for a bond with $100 face value.

    Parameters
    ----------
    settlement : datetime.date
        The bond's settlement date.
    maturity : datetime.date
        The bonds's maturity date (when it expires).
    rate : float
        The bond's annual coupon rate.
    price_ : float
        The bond's quoted clean price.
    redemption : float
        The security's redemption value per $100 face value.
    frequency : int
        The number of coupon payments per year.
    basis : int [optional]
        The type of day count basis to use.
            - 0 [default] : US (NASD) 30/360
            - 1 : Actual/Actual
            - 2 : Actual/360
            - 3 : Actual/365
            - 4 : European 30/360

    Returns
    -------
    float
        The annualized per-period bond yield.
    """

    _PAR = 100
    _coupon_dates = coupon_dates(settlement=settlement, maturity=maturity, frequency=frequency)

    # Special case for when no remaining coupons to be paid (only principal)
    if len(_coupon_dates) == 1:
        _day_count_factor = day_count_factor(start=issue, end=settlement, basis=basis, next_=first_interest, freq=frequency)
        return ((redemption/100 + rate/frequency) - (_PAR/100 + (_day_count_factor * rate))) / (_PAR/100 + (_day_count_factor * rate)) * (frequency / (1 - _day_count_factor * frequency))
    else: # Use optimization to identify the yield that matches quoted clean price
        return scipy.optimize.newton(func=lambda y : price(settlement=settlement, maturity=maturity, rate=rate, yld=y, redemption=redemption, frequency=frequency, basis=basis) - price_, x0=rate, tol=0.0000001, maxiter=100)
